from itertools import cycle
from ..utils.lazy import lazy_property
from ..plugins import PluginBase


class Plain(PluginBase):

    escseq = "\x1b["
    endseq = "\x1b[0m"

    def pride(self, string):
        return "{0}{1}m{2}{3}".format(
            self.escseq, self.colors.next(), string, self.endseq
        )

    def failure(self, string):
        return "{0}31m{1}{2}".format(
            self.escseq, string, self.endseq
        )

    def stack(self, string):
        return "{0}36m{1}{2}".format(
            self.escseq, string, self.endseq
        )

    def generate_colors(self):
        return range(31, 37)

    @lazy_property
    def colors(self):
        return cycle(self.generate_colors())


from nose.plugins import Plugin

class PluginShim(Plugin):

    def beforeTest(self, test):
        self.before_test(test)

    def afterTest(self, test):
        self.after_test(test)

    def prepareTestResult(self, result):
        self.prepare_test_result(result)

    def setOutputStream(self, stream):
        self.set_output_stream(stream)

    def addFailure(self, test, err):
        self.add_failure(test, err)

    def addError(self, test, err):
        self.add_error(test, err)

    def addSuccess(self, test):
        self.add_success(test)

    def addSkip(self, test=None, err=None):
        self.add_skip(test, err)



def lazy_property(fn):
    attr_name = '_lazy_' + fn.__name__

    @property
    def __lazy_property__(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)
    return __lazy_property__



from re import match
from os import getcwd
from os.path import commonprefix, relpath


class Traceback(object):

    def __init__(self, error):
        splitted_lines = error.splitlines()
        self.error_lines = iter(splitted_lines)

        self.assertion_message = splitted_lines[-1]
        self.formatted_lines = []

        self.skip_first_line()

    def skip_first_line(self):
        self.error_lines.next()

    def skip_a_line(self):
        self.error_lines.next()

    def format_line(self):
        matches = match(
            self.source_file_matcher, unicode(self.error_lines.next())
        )

        if matches:
            path, no, method = matches.groups()
            project_path = self.intersect(path)
            if not project_path is path:
                self.formatted_lines.append(
                    "# {0}:{1}:in {2}".format(project_path, no, method
                ))

    def report(self):
        try:
            while True:
                self.format_line()
        except StopIteration:
            return self.formatted_lines

    @staticmethod
    def intersect(path):
        if path.startswith(getcwd()):
            return relpath(path, commonprefix([getcwd(), path]))
        else:
            return path

    @property
    def source_file_matcher(self):
        return r"^  File \"(.+)\", line (\d+), in (.*)$"




class NullStream(object):

    def __init__(self, stream):
        self.stream = stream

    def __getattr__(self, name):
        return getattr(self.stream, name)

    def write(self, *args):
        return

    def writeln(self, *args):
        return self.stream.writeln(*args)


from re import findall
from os import environ
from .formatters.fabulous import Fabulous
from .formatters.plain import Plain


Nosepride = Plain
if findall(r'^xterm|-256color$', environ.get("TERM")):
    Nosepride = Fabulous



from .utils.traceback import Traceback


class FailureReport(object):

    def __init__(self, formatter, result):
        result.printErrors = self.print_errors
        result.printSummary = self.print_summary

        self.formatter = formatter
        self.result = result

    def print_error_header(self, index, test):
        self.formatter.output("  {0}) {1}".format(
            unicode(index + 1), unicode(test)
        ))

    def print_blank_lines(self, count):
        [self.formatter.output("\n") for _ in range(count)]

    def next_failure_message(self):
        return self.formatter.get_next_failed_expectation()

    def print_failure_message(self):
        for line in self.next_failure_message().splitlines():
            self.formatter.output("     {0}\n".format(
                self.formatter.failure(line)
            ))

    def print_traceback(self, error):
        for line in Traceback(error).report():
            self.formatter.output("     {0}\n".format(
                self.formatter.stack(line)
            ))

    def print_failure(self, error):
        self.print_failure_message()
        self.print_traceback(error)

    def print_error(self, error, index, test):
        self.print_error_header(index, test)
        self.print_blank_lines(2)
        self.print_failure(error)
        self.print_blank_lines(1)

    def callback_plugins_report(self):
        self.result.config.plugins.report(self.result.stream)

    def print_errors(self):
        self.print_blank_lines(2)
        errors = self.result.errors + self.result.failures

        if errors:
            self.formatter.output("Failures:")
            self.print_blank_lines(2)
            for index, (test, error) in enumerate(errors):
                self.print_error(error, index, test)

        self.callback_plugins_report()

    def print_colored_finale(self, finale):
        colored_finale = []
        for char in finale:
            colored_finale.append(self.formatter.pride(char))
        return colored_finale

    def print_summary(self, start, stop):
        finale = "Ran %s fabulous tests in %.4f seconds" % (
            self.result.testsRun, stop - start
        )

        colored_finale = self.print_colored_finale(finale)

        self.formatter.output("".join(colored_finale))
        self.print_blank_lines(2)



from .utils.shims import PluginShim
from .streams import NullStream
from .reports import FailureReport


# Plugin interface methods
# https://nose.readthedocs.org/en/latest/plugins/interface.html
class PluginBase(PluginShim):

    score = 199
    name = 'nosepride'
    enabled = True

    def __init__(self):
        super(PluginBase, self).__init__()
        self.stream = None
        self.result = None
        self.failure_report = None
        self.expectation_iterator = None
        self.running_test = False
        self.failed_expectations = []

    def options(self, parser, env):
        parser.add_option(
            "--fabulous-off",
            action="store_false",
            default=True,
            dest="disabled",
            help="disable colour output"
        )

    def failure(self, string):
        raise NotImplementedError(
            "Please provide implementation for failure"
        )

    def pride(self, string):
        raise NotImplementedError(
            "Please provide implementation for pride"
        )

    def stack(self, string):
        raise NotImplementedError(
            "Please provide implementation for pride"
        )

    def get_next_failed_expectation(self):
        if not self.expectation_iterator:
            self.expectation_iterator = iter(self.failed_expectations)
        try:
            return self.expectation_iterator.next()
        except StopIteration:
            return ""

    def record_error(self, test, err):
        self.failed_expectations.append("{0}: {1}".format(
            err[0].__name__, unicode(err[1])
        ))

    def begin(self):
        self.running_test = False

    def before_test(self, test):
        self.running_test = True

    def after_test(self, test):
        if self.running_test:
            self.add_skip()

    def configure(self, options, conf):
        if not options.disabled:
            self.enabled = False

    def prepare_test_result(self, result):
        result.stream = NullStream(result.stream)
        self.failure_report = FailureReport(self, result)
        return result

    def set_output_stream(self, stream):
        self.stream = stream

    def add_failure(self, test, err):
        self.output(self.pride("!"))
        self.record_error(test, err)

    def add_error(self, test, err):
        self.output(self.pride("x"))
        self.record_error(test, err)

    def add_success(self, test):
        self.output(self.pride("."))

    def add_skip(self, test=None, err=None):
        self.output(self.pride("*"))

    def output(self, string):
        self.stream.write(string)
        self.running_test = False



from math import sin, pi
from .plain import Plain


class Fabulous(Plain):

    # colors calculation stolen from Minitest's Pride Plugin
    # https://github.com/seattlerb/minitest
    def calculate_color(self, n):
        n *= 1.0 / 6
        r = int(3 * sin(n) + 3)
        g = int(3 * sin(n + 2 * pi/3) + 3)
        b = int(3 * sin(n + 4 * pi/3) + 3)
        return 36 * r + 6 * g + b + 16

    def pride(self, string):
        return "{0}38;5;{1}m{2}{3}".format(
            self.escseq, self.colors.next(), string, self.endseq
        )

    def generate_colors(self):
        return map(self.calculate_color, range(0, 6 * 7 - 1))



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name="nosepride",
    author="Hugo Bastien",
    author_email="hugobast@gmail.com",
    url="https://github.com/hugobast/nosepride.git",
    version="0.0.7",
    packages=[
        "nosepride",
        "nosepride/formatters",
        "nosepride/utils"
    ],
    tests_require=["mock"],
    install_requires=[
        "nose",
        "setuptools"
    ],
    test_suite='tests',
    license="MIT License",
    description="Fabulous colors for nosetests",
    long_description=open("README.txt").read(),
    entry_points = {
        'nose.plugins.0.10': [
            'nosepride = nosepride:Nosepride'
        ]
    }
)
